#define NUM_FMU_VARIABLES {numFmuVariables}

#define NUM_ONNX_INPUTS {numOnnxInputs}
#define NUM_ONNX_FMU_INPUTS {numOnnxFmuInputs}
#define NUM_ONNX_OUTPUTS {numOnnxOutputs}
#define NUM_ONNX_FMU_OUTPUTS {numOnnxFmuOutputs}
#define NUM_ONNX_STATES {numOnnxStates}
#define NUM_ONNX_STATES_OUTPUTS {numOnnxStatesOutputs}
#define NUM_ONNX_STATE_INIT {numOnnxStateInit}

#define ONNX_INPUT_VALUE_REFERENCES {onnxInputValueReferences}
#define ONNX_OUTPUT_VALUE_REFERENCES {onnxOutputValueReferences}
#define ONNX_STATE_OUTPUT_INDEXES {onnxStateOutputIndexes}
#define ONNX_STATE_INIT_VALUE_REFERENCES {onnxStateInitValueReferences}

#define ONNX_USE_TIME_INPUT {onnxUsesTime}

#define ONNX_INPUT_NAME "{onnxInputName}"
#define ONNX_OUTPUT_NAME "{onnxOutputName}"
#define ONNX_STATE_NAME "{onnxStatesName}"
#define ONNX_TIME_INPUT_NAME "{onnxTimeInputName}"
#define ONNX_FILENAME L"{onnxFileName}"